import logging

import numpy as np
import parmed as pmd
import pytraj as pt

from paprika import utils

logger = logging.getLogger(__name__)


class DAT_restraint(object):
    """
    Distance or angle or torsion restraints on atoms in the simulation.
    """

    @property
    def instances(self):
        """A list of ``DAT_restraints`` that have been initialized.

        .. note ::

        This should never be called directly and ought to be private.
        """
        return self._instances

    @instances.setter
    def instances(self, value):
        self._instances = value

    @property
    def topology(self):
        """The topology file used to initialize the restraints. This is often a PDB file."""
        return self._topology

    @topology.setter
    def topology(self, value):
        self._topology = value

    @property
    def mask1(self):
        """The first atom mask used for the restraint."""
        return self._mask1

    @mask1.setter
    def mask1(self, value):
        self._mask1 = value

    @property
    def mask2(self):
        """The second atom mask used for the restraint."""
        return self._mask2

    @mask2.setter
    def mask2(self, value):
        self._mask2 = value

    @property
    def mask3(self):
        """The third atom mask used for the restraint. Only two atom masks are required."""
        return self._mask3

    @mask3.setter
    def mask3(self, value):
        self._mask3 = value

    @property
    def mask4(self):
        """The fourth atom mask used for the restraint. Only two atom masks are required."""
        return self._mask4

    @mask4.setter
    def mask4(self, value):
        self._mask4 = value

    @property
    def custom_restraint_values(self):
        """In the case a non-harmonic restraint is desired, the pre-calculated values (r1, r2, and so on) can be
        overridden with ones from this dictionary. These values will be directly written to the AMBER restraint file,
        so the keys must be valid AMBER keywords.

        Specifically, for target distance, R:

        • R < r1 Linear, with the slope of the "left-hand" parabola at the point R = r1.
        • r1 <= R < r2 Parabolic, with restraint energy k2(R − r2)².
        • r2 <= R < r3 E = 0.
        • r3 <= R < r4 Parabolic, with restraint energy k3(R − r3)².
        • r4 <= R Linear, with the slope of the "right-hand" parabola at the point R = r4.

        In the case of AMBER18, this is covered in section 25.1 of the manual.

        For example, a flat bottom restraint can be specified by setting ``r2`` not equal to ``r3``.

        >>> from paprika.restraints import DAT_restraint
        >>> this = DAT_restraint
        >>> this.custom_restraint_values["r2"] = 0.0
        >>> this.custom_restraint_values["r3"] = 12.0

        The values in this dictionary use default AMBER units for distances and force constants.

        This attribute has **no effect** for OpenMM.
        """
        return self._custom_restraint_values

    @custom_restraint_values.setter
    def custom_restraint_values(self, value):
        self._custom_restraint_values = value

    @property
    def auto_apr(self):
        """If ``True``, (1) the force constant during the pulling phase will be set to the final force constant from
        the attach phase, and (2) the initial target during the pulling phase will be equal to the final target from
        the attach phase. If ``False``, these values must be set manually.
        """
        return self._auto_apr

    @auto_apr.setter
    def auto_apr(self, value):
        if not isinstance(value, bool):
            raise TypeError("The ``auto_apr`` attribute is a boolean.")
        self._auto_apr = value

    @property
    def continuous_apr(self):
        """
        If ``True``, (1) the final window of the attach phase is used as the first window of the pull phase, and (2) the
        final window of the pull phase is used as the first window of the release phase.
        """
        return self._continuous_apr

    @continuous_apr.setter
    def continuous_apr(self, value):
        if not isinstance(value, bool):
            raise TypeError("The ``continuous_apr`` attribute is a boolean.")
        self._continuous_apr = value

    @property
    def attach(self):
        """
        Dictionary specifying the APR parameters during the attach phase. The dictionary keys are as follows:

            - ``target``             : The target value for the restraint (mandatory)
            - ``fc_initial``         : The initial force constant (optional)
            - ``fc_final``           : The final force constant (optional)
            - ``num_windows``        : The number of windows (optional)
            - ``fc_increment``       : The force constant increment (optional)
            - ``fraction_increment`` : The percentage of the force constant increment (optional)
            - ``fraction_list``      : The list of force constant percentages (optional)
            - ``fc_list``            : The list of force constants (will be created if not given)

        .. note ::
            This is fragile and this could be hardened by making these ``ENUM``s and doing much more type-checking.
        """

        return self._attach

    @attach.setter
    def attach(self, value):
        self._attach = value

    @property
    def pull(self):
        """
        Dictionary specifying the APR parameters during the pull phase. The dictionary keys are as follows:

            - ``fc``                 : The force constant for the restraint (mandatory)
            - ``target_initial``     : The initial target value (optional)
            - ``target_final``       : The final target value (optional)
            - ``num_windows``        : The number of windows (optional)
            - ``target_increment``   : The target value increment (optional)
            - ``fraction_increment`` : The percentage of the target value increment (optional)
            - ``fraction_list``      : The list of target value percentages (optional)
            - ``target_list``        : The list of target values (will be created if not given)

        .. note ::
            This is fragile and this could be hardened by making these ``ENUM``s and doing much more type-checking.
        """

        return self._pull

    @pull.setter
    def pull(self, value):
        self._pull = value

    @property
    def release(self):
        """
        Dictionary specifying the APR parameters during the release phase. The dictionary keys are as follows:

            - ``target``             : The target value for the restraint (mandatory)
            - ``fc_initial``         : The initial force constant (optional)
            - ``fc_final``           : The final force constant (optional)
            - ``num_windows``        : The number of windows (optional)
            - ``fc_increment``       : The force constant increment (optional)
            - ``fraction_increment`` : The percentage of the force constant increment (optional)
            - ``fraction_list``      : The list of force constant percentages (optional)
            - ``fc_list``            : The list of force constants (will be created if not given)

        .. note ::
            This is fragile and this could be hardened by making these ``ENUM``s and doing much more type-checking.
        """
        return self._release

    @release.setter
    def release(self, value):
        self._release = value

    @property
    def amber_index(self):
        """
        If ``True``, add 1 to all atom indices.
        """
        return self._amber_index

    @amber_index.setter
    def amber_index(self, value):
        self._amber_index = value

    instances = []

    def __init__(self):

        self._topology = None
        self._mask1 = None
        self._mask2 = None
        self._mask3 = None
        self._mask4 = None

        # These indices will be automatically populated during :meth:`paprika.restraints.DAT_restraint.initialize`.
        self.index1 = None
        self.index2 = None
        self.index3 = None
        self.index4 = None

        self.group1 = False
        self.group2 = False
        self.group3 = False
        self.group4 = False

        self._custom_restraint_values = {}

        self._auto_apr = False
        self._continuous_apr = True

        self._amber_index = False

        self._attach = {
            "target": None,
            "fc_initial": None,
            "fc_final": None,
            "num_windows": None,
            "fc_increment": None,
            "fraction_increment": None,
            "fraction_list": None,
            "fc_list": None,
        }
        self._pull = {
            "fc": None,
            "target_initial": None,
            "target_final": None,
            "num_windows": None,
            "target_increment": None,
            "fraction_increment": None,
            "fraction_list": None,
            "target_list": None,
        }
        self._release = {
            "target": None,
            "fc_initial": None,
            "fc_final": None,
            "num_windows": None,
            "fc_increment": None,
            "fraction_increment": None,
            "fraction_list": None,
            "fc_list": None,
        }

        DAT_restraint.instances.append(self)

    def __eq__(self, other):
        """
        Test whether two ``DAT_restraint`` instances are equivalent.
        """
        self_dictionary = self.__dict__
        other_dictionary = other.__dict__
        for dct in [self_dictionary, other_dictionary]:
            # Skip checking topology.
            # It is difficult to check topology for two reasons.
            # One, there are numerical variances in the numeric parameters.
            # We can use `np.allcose()` but it needs to be done manually on the
            # nested ParmEd AmberParm or Structure classes.
            # Two, sometimes the dummy atoms seem to be encoded with index `-1`.
            # This can happen if the structure used to setup the restraint
            # does not have the dummy atoms already included.
            # Previously, I checked if the topology was an AmberParm, and tried
            # to replace that with the string representation of the file name,
            # but for some reason, that attribute was not always available.
            # Thus, I simply do not compare the `topology` attribute.
            dct["topology"] = None
        logger.debug(self_dictionary)
        logger.debug(other_dictionary)
        keys = set(self_dictionary.keys()) & set(other_dictionary.keys())
        for key in keys:
            if key != "phase":
                assert self_dictionary[key] == other_dictionary[key]
            else:
                for phs in ["attach", "pull", "release"]:
                    for value in ["force_constants", "targets"]:
                        if (
                                self_dictionary["phase"][phs][value] is None
                                and other_dictionary["phase"][phs][value] is None
                        ):
                            continue
                        else:
                            assert np.allclose(
                                self_dictionary["phase"][phs][value],
                                other_dictionary["phase"][phs][value],
                            )
        return True

    @staticmethod
    def _calc_method(phase, restraint_dictionary, method):
        """
        This helper function figures out which values in the restraint dictionary need to be set.

        phase: str
            The restraint phase.
        restraint_dictionary: dict
            The restraint dictionary corresponding whose values will be set.
        method: str
            Which "method" will be used to set the remaining values. The "method" refers to which combination of restraint
            inputs are provided -- for example, initial value, final value, and number of windows or initial value, increment
            size, and number of increments -- which are defined in :meth:`paprika.restraints.DAT_restraint.initialize`.

            This appears to be a ``str`` but should be an ``int``.

        .. note ::
            This could do with some serious sprucing up.

        """

        force_constants = None
        targets = None

        # Attach/Release, Force Constant Method 1
        if phase in ("a", "r") and method == "1":
            force_constants = np.linspace(
                restraint_dictionary["fc_initial"], restraint_dictionary["fc_final"],
                restraint_dictionary["num_windows"]
            )

        # Attach/Release, Force Constant Method 1a
        elif phase in ("a", "r") and method == "1a":
            force_constants = np.linspace(0.0, restraint_dictionary["fc_final"], restraint_dictionary["num_windows"])

        # Attach/Release, Force Constant Method 2
        elif phase in ("a", "r") and method == "2":
            force_constants = np.arange(
                restraint_dictionary["fc_initial"],
                restraint_dictionary["fc_final"]+restraint_dictionary["fc_increment"],
                restraint_dictionary["fc_increment"],
            )

        # Attach/Release, Force Constant Method 2a
        elif phase in ("a", "r") and method == "2a":
            force_constants = np.arange(
                0.0, restraint_dictionary["fc_final"]+restraint_dictionary["fc_increment"],
                restraint_dictionary["fc_increment"]
            )

        # Attach/Release, Force Constant Method 3
        elif phase in ("a", "r") and method == "3":
            force_constants = np.asarray(
                [fraction * restraint_dictionary["fc_final"] for fraction in restraint_dictionary["fraction_list"]]
            )

        # Attach/Release, Force Constant Method 4
        elif phase in ("a", "r") and method == "4":
            fractions = np.arange(
                0, 1.0+restraint_dictionary["fraction_increment"], restraint_dictionary["fraction_increment"]
            )
            force_constants = np.asarray(
                [fraction * restraint_dictionary["fc_final"] for fraction in fractions]
            )

        # Attach/Release, Force Constant Method 5
        elif phase in ("a", "r") and method == "5":
            force_constants = np.asarray(restraint_dictionary["fc_list"])

        # Attach/Release, Target Method
        if phase in ("a", "r"):
            targets = np.asarray([restraint_dictionary["target"]] * len(force_constants))

        # Pull, Target Method 1
        if phase == "p" and method == "1":
            targets = np.linspace(
                restraint_dictionary["target_initial"], restraint_dictionary["target_final"],
                restraint_dictionary["num_windows"]
            )

        # Pull, Target Method 1a
        elif phase == "p" and method == "1a":
            targets = np.linspace(0.0, restraint_dictionary["target_final"], restraint_dictionary["num_windows"])

        # Pull, Target Method 2
        elif phase == "p" and method == "2":
            targets = np.arange(
                restraint_dictionary["target_initial"],
                restraint_dictionary["target_final"]+restraint_dictionary["target_increment"],
                restraint_dictionary["target_increment"],
            )

        # Pull, Target Method 2a
        elif phase == "p" and method == "2a":
            targets = np.arange(
                0.0,
                restraint_dictionary["target_final"]+restraint_dictionary["target_increment"],
                restraint_dictionary["target_increment"],
            )

        # Pull, Target Method 3
        elif phase == "p" and method == "3":
            targets = np.asarray(
                [
                    fraction * restraint_dictionary["target_final"]
                    for fraction in restraint_dictionary["fraction_list"]
                ]
            )

        # Pull, Target Method 4
        elif phase == "p" and method == "4":
            fractions = np.arange(
                0, 1.0+restraint_dictionary["fraction_increment"], restraint_dictionary["fraction_increment"]
            )
            targets = np.asarray(
                [fraction * restraint_dictionary["target_final"] for fraction in fractions]
            )

        # Pull, Target Method 5
        elif phase == "p" and method == "5":
            targets = np.asarray(restraint_dictionary["target_list"])

        # Pull, Force Constant Method
        if phase == "p":
            force_constants = np.asarray([restraint_dictionary["fc"]] * len(targets))

        if force_constants is None and targets is None:
            logger.error("Unsupported Phase/Method: {} / {}".format(phase, method))
            raise Exception("Unexpected phase/method combination passed to _calc_meth")

        return force_constants, targets

    def initialize(self):
        """
        Automatically set remaining force constants and targets.

        Depending on which values are provided for each phase, a different method will
        be used to determine the list of force constants and targets (below).

        For attach and release, a ``target`` value is required and the method is determined if the
        following values are not ``None``:

            - Method 1:   num_windows, fc_initial, fc_final
            - Method 1a:  num_windows, fc_final
            - Method 2:   fc_increment, fc_initial, fc_final
            - Method 2a:  fc_increment, fc_final
            - Method 3:   fraction_list, fc_final
            - Method 4:   fraction_increment, fc_final
            - Method 5:   fc_list

        For pull, a ``fc`` value is required and the method is determined if the
        following values are not ``None``:

            - Method 1:   num_windows, target_initial, target_final
            - Method 1a:  num_windows, target_final
            - Method 2:   target_increment, target_initial, target_final
            - Method 2a:  target_increment, target_final
            - Method 3:   fraction_list, target_final
            - Method 4:   fraction_increment, target_final
            - Method 5:   target_list

        .. note ::
            This is unnecessary overengineering.
        """

        self.phase = {
            "attach": {"force_constants": None, "targets": None},
            "pull": {"force_constants": None, "targets": None},
            "release": {"force_constants": None, "targets": None},
        }
        # ------------------------------------ ATTACH ------------------------------------ #
        logger.debug("Calculating attach targets and force constants...")

        # Temporary variables to improve readability
        force_constants = None
        targets = None

        if (
                self.attach["num_windows"] is not None
                and self.attach["fc_final"] is not None
        ):
            if self.attach["fc_initial"] is not None:
                logger.debug("Attach, Method #1")
                force_constants, targets = self._calc_method("a", self.attach, "1")
            else:
                logger.debug("Attach, Method #1a")
                force_constants, targets = self._calc_method("a", self.attach, "1a")

        elif (
                self.attach["fc_increment"] is not None
                and self.attach["fc_final"] is not None
        ):
            if self.attach["fc_initial"] is not None:
                logger.debug("Attach, Method #2")
                force_constants, targets = self._calc_method("a", self.attach, "2")
            else:
                logger.debug("Attach, Method #2a")
                force_constants, targets = self._calc_method("a", self.attach, "2a")

        elif (
                self.attach["fraction_list"] is not None
                and self.attach["fc_final"] is not None
        ):
            logger.debug("Attach, Method #3")
            force_constants, targets = self._calc_method("a", self.attach, "3")

        elif (
                self.attach["fraction_increment"] is not None
                and self.attach["fc_final"] is not None
        ):
            logger.debug("Attach, Method #4")
            force_constants, targets = self._calc_method("a", self.attach, "4")

        elif self.attach["fc_list"] is not None:
            logger.debug("Attach, Method #5")
            force_constants, targets = self._calc_method("a", self.attach, "5")

        elif all(v is None for k, v in self.attach.items()):
            logger.debug("No restraint info set for the attach phase! Skipping...")

        else:
            logger.error(
                "Attach restraint input did not match one of the supported methods..."
            )
            for k, v in self.attach.items():
                logger.debug("{} = {}".format(k, v))
            raise Exception(
                "Attach restraint input did not match one of the supported methods..."
            )

        if force_constants is not None and targets is not None:
            self.phase["attach"]["force_constants"] = force_constants
            self.phase["attach"]["targets"] = targets

        # ------------------------------------ PULL ------------------------------------ #
        logger.debug("Calculating pull targets and force constants...")

        force_constants = None
        targets = None

        if self.auto_apr and self.pull["target_final"] is not None:
            self.pull["fc"] = self.phase["attach"]["force_constants"][-1]
            self.pull["target_initial"] = self.phase["attach"]["targets"][-1]

        if (
                self.pull["num_windows"] is not None
                and self.pull["target_final"] is not None
        ):
            if self.pull["target_initial"] is not None:
                logger.debug("Pull, Method #1")
                force_constants, targets = self._calc_method("p", self.pull, "1")
            else:
                logger.debug("Pull, Method #1a")
                force_constants, targets = self._calc_method("p", self.pull, "1a")

        elif (
                self.pull["target_increment"] is not None
                and self.pull["target_final"] is not None
        ):
            if self.pull["target_initial"] is not None:
                logger.debug("Pull, Method #2")
                force_constants, targets = self._calc_method("p", self.pull, "2")
            else:
                logger.debug("Pull, Method #2a")
                force_constants, targets = self._calc_method("p", self.pull, "2a")

        elif (
                self.pull["fraction_list"] is not None
                and self.pull["target_final"] is not None
        ):
            logger.debug("Pull, Method #3")
            force_constants, targets = self._calc_method("p", self.pull, "3")

        elif (
                self.pull["fraction_increment"] is not None
                and self.pull["target_final"] is not None
        ):
            logger.debug("Pull, Method #4")
            force_constants, targets = self._calc_method("p", self.pull, "4")

        elif self.pull["target_list"] is not None:
            logger.debug("Pull, Method #5")
            force_constants, targets = self._calc_method("p", self.pull, "5")

        elif all(v is None for k, v in self.pull.items()):
            logger.debug("No restraint info set for the pull phase! Skipping...")

        else:
            logger.error(
                "Pull restraint input did not match one of the supported methods..."
            )
            for k, v in self.pull.items():
                logger.debug("{} = {}".format(k, v))
            raise Exception(
                "Pull restraint input did not match one of the supported methods..."
            )

        if force_constants is not None and targets is not None:
            self.phase["pull"]["force_constants"] = force_constants
            self.phase["pull"]["targets"] = targets

        # ------------------------------------ RELEASE ------------------------------------ #
        logger.debug("Calculating release targets and force constants...")

        force_constants = None
        targets = None

        # I don't want auto_apr to make release restraints, unless I'm sure the user wants them.
        # I'm gonna assume that specifying self.attach['fc_final'] indicates you want it,
        # although this weakens the whole purpose of auto_apr.

        if self.auto_apr and self.release["fc_final"] is not None:
            self.release["target"] = self.phase["pull"]["targets"][-1]
            for key in [
                "fc_final",
                "fc_initial",
                "num_windows",
                "fc_increment",
                "fraction_increment",
                "fraction_list",
                "fc_list",
            ]:
                if self.attach[key] is not None and self.release[key] is None:
                    self.release[key] = self.attach[key]

        if (
                self.release["num_windows"] is not None
                and self.release["fc_final"] is not None
        ):
            if self.release["fc_initial"] is not None:
                logger.debug("Release, Method #1")
                force_constants, targets = self._calc_method("r", self.release, "1")
            else:
                logger.debug("Release, Method #1a")
                force_constants, targets = self._calc_method("r", self.release, "1a")

        elif (
                self.release["fc_increment"] is not None
                and self.release["fc_final"] is not None
        ):
            if self.release["fc_initial"] is not None:
                logger.debug("Release, Method #2")
                force_constants, targets = self._calc_method("r", self.release, "2")
            else:
                logger.debug("Release, Method #2a")
                force_constants, targets = self._calc_method("r", self.release, "2a")

        elif (
                self.release["fraction_list"] is not None
                and self.release["fc_final"] is not None
        ):
            logger.debug("Release, Method #3")
            force_constants, targets = self._calc_method("r", self.release, "3")

        elif (
                self.release["fraction_increment"] is not None
                and self.release["fc_final"] is not None
        ):
            logger.debug("Release, Method #4")
            force_constants, targets = self._calc_method("r", self.release, "4")

        elif self.release["fc_list"] is not None:
            logger.debug("Release, Method #5")
            force_constants, targets = self._calc_method("r", self.release, "5")

        elif all(v is None for k, v in self.release.items()):
            logger.debug("No restraint info set for the release phase! Skipping...")

        else:
            logger.error(
                "Release restraint input did not match one of the supported methods..."
            )
            for k, v in self.release.items():
                logger.debug("{} = {}".format(k, v))
            raise Exception(
                "Release restraint input did not match one of the supported methods..."
            )

        if force_constants is not None and targets is not None:
            self.phase["release"]["force_constants"] = force_constants
            self.phase["release"]["targets"] = targets

        # ----------------------------------- WINDOWS ------------------------------------ #

        for phase in ["attach", "pull", "release"]:
            if self.phase[phase]["targets"] is not None:
                window_count = len(self.phase[phase]["targets"])
                logger.debug("Number of {} windows = {}".format(phase, window_count))
            else:
                logger.debug(
                    "This restraint will be skipped in the {} phase".format(phase)
                )

        # ---------------------------------- ATOM MASKS ---------------------------------- #
        logger.debug("Assigning atom indices...")
        self.index1 = utils.index_from_mask(self.topology, self.mask1, self.amber_index)
        self.index2 = utils.index_from_mask(self.topology, self.mask2, self.amber_index)
        if self.mask3:
            self.index3 = utils.index_from_mask(
                self.topology, self.mask3, self.amber_index
            )
        else:
            self.index3 = None
        if self.mask4:
            self.index4 = utils.index_from_mask(
                self.topology, self.mask4, self.amber_index
            )
        else:
            self.index4 = None
        # If any `index` has more than one atom, mark it as a group restraint.
        if self.mask1 and len(self.index1) > 1:
            self.group1 = True

        if self.mask2 and len(self.index2) > 1:
            self.group2 = True

        if self.mask3 and len(self.index3) > 1:
            self.group3 = True

        if self.mask4 and len(self.index4) > 1:
            self.group4 = True


def static_DAT_restraint(
        restraint_mask_list,
        num_window_list,
        ref_structure,
        force_constant,
        continuous_apr=True,
        amber_index=False,
):
    """
    Create a restraint whose value does not change during a calculation.

    Parameters
    ----------
    restraint_mask_list: list
        A list of masks for which this restraint applies.
    num_window_list: list
        A list of windows during which this restraint will be applied, which should be in the form: [attach windows,
        pull windows, release windows].
    ref_structure: Path-like
        The reference structure that is used to determine the initial, **static** value for this restraint.
    force_constant: float
        The force constant for this reestraint.
    continuous_apr: bool
        Whether this restraint uses ``continuous_apr``. This must be consistent with existing restraints.
    amber_index: bool
        Whether the atom indices for the restraint should be AMBER-style (+1) or not.

    Returns
    -------
    rest: ``DAT_restraint``
        A static restraint.

    """

    reference_trajectory = pt.iterload(ref_structure, traj=True)

    # Check num_window_list
    if len(num_window_list) != 3:
        raise ValueError(
            "The num_window_list needs to contain three integers corresponding to the number of windows in the "
            "attach, pull, and release phase, respectively "
        )

    rest = DAT_restraint()
    rest.continuous_apr = continuous_apr
    rest.amber_index = amber_index
    rest.topology = pmd.load_file(ref_structure, structure=True)
    rest.mask1 = restraint_mask_list[0]
    rest.mask2 = restraint_mask_list[1]
    if len(restraint_mask_list) >= 3:
        rest.mask3 = restraint_mask_list[2]
    if len(restraint_mask_list) == 4:
        rest.mask4 = restraint_mask_list[3]

    # Target value
    mask_string = " ".join(restraint_mask_list)
    if len(restraint_mask_list) == 2:
        # Distance restraint
        target = pt.distance(reference_trajectory, mask_string, image=True)[0]
    elif len(restraint_mask_list) == 3:
        # Angle restraint
        target = pt.angle(reference_trajectory, mask_string)[0]
    elif len(restraint_mask_list) == 4:
        # Dihedral restraint
        target = pt.dihedral(reference_trajectory, mask_string)[0]
    else:
        raise IndexError(
            f"The number of masks -- {len(restraint_mask_list)} -- is not 2, 3, or 4 and thus is not one of the "
            f"supported types: distance, angle, or dihedral."
        )

    # Attach phase
    if num_window_list[0] is not None and num_window_list[0] != 0:
        rest.attach["target"] = target
        rest.attach["fc_initial"] = force_constant
        rest.attach["fc_final"] = force_constant
        rest.attach["num_windows"] = num_window_list[0]

    # Pull phase
    if num_window_list[1] is not None and num_window_list[1] != 0:
        rest.pull["fc"] = force_constant
        rest.pull["target_initial"] = target
        rest.pull["target_final"] = target
        rest.pull["num_windows"] = num_window_list[1]

    # Release phase
    if num_window_list[2] is not None and num_window_list[2] != 0:
        rest.release["target"] = target
        rest.release["fc_initial"] = force_constant
        rest.release["fc_final"] = force_constant
        rest.release["num_windows"] = num_window_list[2]

    rest.initialize()

    return rest


def check_restraints(restraint_list, create_window_list=False):
    """
    Do basic tests to ensure a list of ``DAT_restraint``s are consistent.
    This function is also overloaded to create window lists as well, which is not ideal.

    Parameters
    ----------
    restraint_list: list
        A list of restraints.
    create_window_list: bool
        Whether to use the restraints to create windows for the calculation.

    """

    if all(restraint.continuous_apr is True for restraint in restraint_list):
        logger.debug('All restraints are "continuous_apr" style.')
        all_continuous_apr = True
    elif all(restraint.continuous_apr is False for restraint in restraint_list):
        logger.debug('All restraints are not "continuous_apr" style.')
        all_continuous_apr = False
    else:
        raise ValueError("All restraints must have the same setting for ``.continuous_apr``.")

    window_list = []
    phases = ["attach", "pull", "release"]
    for phase in phases:
        win_counts = []
        for restraint in restraint_list:
            if restraint.phase[phase]["targets"] is not None:
                win_counts.append(len(restraint.phase[phase]["targets"]))
            else:
                win_counts.append(0)
        max_count = np.max(win_counts)

        if max_count > 999:
            logger.info("Window name zero padding only applied up to 999.")

        # For each restraint, make sure the number of windows is either 0 (the restraint
        # is not active) or equal to the maximum number of windows for any
        # restraint.
        if all(count == 0 or count == max_count for count in win_counts):
            if max_count > 0:
                # `continuous_apr` during attach means that the final attach window
                # should be skipped and replaced with `p000`. `continuous_apr` during
                # release means that `r000` should be skipped and replaced with the
                # final pull window.

                if phase == "attach" and all_continuous_apr:
                    window_list += [
                        phase[0]+str("{:03.0f}".format(val))
                        for val in np.arange(0, max_count-1, 1)
                    ]
                elif phase == "attach" and not all_continuous_apr:
                    window_list += [
                        phase[0]+str("{:03.0f}".format(val))
                        for val in np.arange(0, max_count, 1)
                    ]
                elif phase == "pull":
                    window_list += [
                        phase[0]+str("{:03.0f}".format(val))
                        for val in np.arange(0, max_count, 1)
                    ]
                elif phase == "release" and all_continuous_apr:
                    window_list += [
                        phase[0]+str("{:03.0f}".format(val))
                        for val in np.arange(1, max_count, 1)
                    ]
                elif phase == "release" and not all_continuous_apr:
                    window_list += [
                        phase[0]+str("{:03.0f}".format(val))
                        for val in np.arange(0, max_count, 1)
                    ]
        else:
            logger.error(
                "Restraints have unequal number of windows during the {} phase.".format(
                    phase
                )
            )
            logger.debug("Window counts for each restraint are as follows:")
            logger.debug(win_counts)
            raise Exception(
                "Restraints have unequal number of windows during the {} "
                "phase.".format(phase)
            )

    logger.info("Restraints appear to be consistent")

    if create_window_list:
        return window_list


def create_window_list(restraint_list):
    """
    Create list of APR windows. Runs everything through check_restraints because
    we need to do that.
    """

    return check_restraints(restraint_list, create_window_list=True)
